from requests.auth import AuthBase


class BearerAuth(AuthBase):

    def __init__(self, token: str):
        self.__token = token

    def __call__(self, r):
        r.headers['Authorization'] = 'Bearer {}'.format(self.__token)
        return r


class URLs:

    # Auth
    @classmethod
    def authenticate(cls):
        """Returns the url to authenticate a user"""
        return 'https://authserver.mojang.com/authenticate'

    @classmethod
    def refresh(cls):
        """Returns the url to refresh an access token"""
        return 'https://authserver.mojang.com/refresh'

    @classmethod
    def validate(cls):
        """Returns the urls to validate an access token"""
        return 'https://authserver.mojang.com/validate'

    @classmethod
    def invalidate(cls):
        """Returns the urls to invalidate an access token"""
        return 'https://authserver.mojang.com/invalidate'

    @classmethod
    def signout(cls):
        """Returns the urls to signout a user"""
        return 'https://authserver.mojang.com/signout'

    ## Security
    @classmethod
    def verify_ip(cls):
        """Returns the url to check and verify user IP"""
        return 'https://api.mojang.com/user/security/location'

    @classmethod
    def get_challenges(cls):
        """Returns the url to get the challenges for user IP verification"""
        return 'https://api.mojang.com/user/security/challenges'




"""
Pymojang module
---------------

Pymojang is a full wrapper around the [Mojang API](https://wiki.vg/Mojang_API) 
and [Mojang Authentication API](https://wiki.vg/Authentication)

Checkout the [`documentation`](https://pymojang.readthedocs.io/en/latest/)
"""
from .account import get_uuid, get_uuids, names, status, user, connect



import struct
from typing import IO, Optional


class VarInt:

    @classmethod
    def write(cls, buffer: IO, value: int):
        while True:
            byte = value & 0x7f
            value >>= 7

            if value > 0:
                byte |= 0x80

            buffer.write(struct.pack('B', byte))
            
            if value == 0:
                break


    @classmethod
    def read(cls, buffer: IO) -> int:
        val = 0

        for i in range(5):
            byte = buffer.read(1)
            if len(byte) == 0:
                break

            val |= (ord(byte) & 0x7f) << (7*i)
            
            if ord(byte) & 0x80 == 0:
                break

        return val


class String:

    @classmethod
    def write(cls, buffer: IO, value: str, encoding: Optional[str] = 'utf-8'):
        value = value.encode(encoding)
        VarInt.write(buffer, len(value))
        buffer.write(value)


    @classmethod
    def read(cls, buffer: IO, encoding: Optional[str] = 'utf-8') -> str:
        length = VarInt.read(buffer)
        return buffer.read(length).decode(encoding)



import random
import socket
import struct
from typing import Tuple


class Packets:

    def __init__(self, sock: socket.socket):
        self.__sock = sock


    def _new_id(self) -> int:
        return random.randint(0,2**31)


    def send(self, _type: int, payload: str) -> Tuple[int, int]:
        packet_id = self._new_id()
        packet = struct.pack(
            f'<iii{len(payload) + 2}s',
            10 + len(payload),
            packet_id,
            _type,
            payload.encode('ascii') + b'\00'
        )
        sent = self.__sock.send(packet)
        return packet_id, sent


    def recv(self) -> Tuple[int, int, bytes, int]:
        with self.__sock.makefile('rb') as buffer:
            size = int.from_bytes(buffer.read(4), 'little')

            # Read fully
            data = b''
            while len(data) < size:
                _data = buffer.read(size - len(data))
                if _data:
                    data += _data
        
        packet_id, _type = struct.unpack('<ii', data[:8])
        return packet_id, _type, data[8:], size - 8



from typing import NamedTuple


## Authentication 
class AuthenticationInfo(NamedTuple):
    """
    Attributes:
        access_token (str): The session's access token
        client_token (str): The session's client token
        uuid (str): The uuid of the player
        name (str): The name of the player
        legacy (bool): Wether the account has migrated
        demo (bool): Wether the account is demo 
    """
    access_token: str
    client_token: str
    uuid: str
    name: str
    legacy: bool
    demo: bool

## Security
class ChallengeInfo(NamedTuple):
    """
    Attributes:
        id (int): The id of the challenge
        challenge (str): The challenge to complete
    """
    id: int
    challenge: str



from typing import Optional

import requests

from ...exceptions import *
from ..structures.auth import AuthenticationInfo
from ..utils.auth import URLs


def authenticate(username: str, password: str, client_token: Optional[str] = None) -> AuthenticationInfo:
    """Authenticate a user with name and password

    Args:
        username (str): The username of email if account is not legacy
        password (str): The user password
        client_token (str, optional): The client token to use (default to None)

    Returns:
        AuthenticationInfo

    Raises:
        CredentialsError: If username and password are invalid
        PayloadError: If credentials are not formated correctly

    Example:

        ```python
        from mojang.account.auth import yggdrasil

        auth_info = yggdrasil.authenticate('USERNAME_OR_EMAIL','PASSWORD')
        print(auth_info)
        ```
        ```
        AuthenticationInfo(access_token='ACCESS_TOKEN', client_token='CLIENT_TOKEN', uuid='...', name='...', legacy=False, demo=False)
        ```
    """
    payload = {
        'username': username,
        'password': password,
        'clientToken': client_token,
        'agent': {
            'name': 'Minecraft',
            'version': 1
        }
    }
    response = requests.post(URLs.authenticate(), json=payload)
    data = handle_response(response, PayloadError, CredentialsError)

    _dict = {
        'access_token': data['accessToken'],
        'client_token': data['clientToken'],
        'uuid': data['selectedProfile']['id'],
        'name': data['selectedProfile']['name'],
        'legacy': data['selectedProfile'].get('legacy', False),
        'demo': not data['selectedProfile'].get('paid', True)
    }
    return AuthenticationInfo(**_dict)

def refresh(access_token: str, client_token: str) -> AuthenticationInfo:
    """Refresh an invalid access token
    
    Args:
        access_token (str): The access token to refresh
        client_token (str): The client token used to generate the access token
    
    Returns:
        AuthenticationInfo

    Raises:
        TokenError: If client token is not the one used to generate the access token
        PayloadError: If the tokens are not formated correctly

    Example:

        ```python
        from mojang.account.auth import yggdrasil

        refresh_info = yggdrasil.refresh('CURRENT_ACCESS_TOKEN','CLIENT_TOKEN')
        print(refresh_info)
        ```
        ```
        AuthenticationInfo(access_token='NEW_ACCESS_TOKEN', client_token='CLIENT_TOKEN', uuid='...', name='...', legacy=False, demo=False)
        ```
    """
    payload = {
        'accessToken': access_token,
        'clientToken': client_token
    }
    response = requests.post(URLs.refresh(), json=payload)
    data = handle_response(response, PayloadError, TokenError)

    _dict = {
        'access_token': data['accessToken'],
        'client_token': data['clientToken'],
        'uuid': data['selectedProfile']['id'],
        'name': data['selectedProfile']['name'],
        'legacy': data['selectedProfile'].get('legacy', False),
        'demo': not data['selectedProfile'].get('paid', True)
    }
    return AuthenticationInfo(**_dict)

def validate(access_token: str, client_token: str):
    """Validate an access token
    
    Args:
        access_token (str): The access token to validate
        client_token (str): The client token used to generate the access token
    
    Raises:
        TokenError: If client token is not the one used to generate the access token
        PayloadError: If the tokens are not formated correctly
    
    Example:

        ```python
        from mojang.account.auth import yggdrasil

        yggdrasil.validate('CURRENT_ACCESS_TOKEN','CLIENT_TOKEN')
        ```
    """
    payload = {
        'accessToken': access_token,
        'clientToken': client_token
    }
    response = requests.post(URLs.validate(), json=payload)
    handle_response(response, PayloadError, TokenError)

def signout(username: str, password: str):
    """Signout user with name and password
    
    Args:
        username (str): The username or email if account is not legacy
        password (str): The user password
    
    Raises:
        CredentialsError: If username and password are invalid
        PayloadError: If credentials are not formated correctly

    Example:

        ```python
        from mojang.account.auth import yggdrasil
        
        yggdrasil.signout('USERNAME_OR_EMAIL','PASSWORD')
        ```
    """
    payload = {
        'username': username,
        'password': password
    }
    response = requests.post(URLs.signout(), json=payload)
    handle_response(response, PayloadError, CredentialsError)

def invalidate(access_token: str, client_token: str):
    """Invalidate an access token
    
    Args:
        access_token (str): The access token to invalidate
        client_token (str): The client token used to generate the access token
    
    Raises:
        TokenError: If client token is not the one used to generate the access token
        PayloadError: If the tokens are not formated correctly
    
    Example:

        ```python
        from mojang.account.auth import yggdrasil
        
        yggdrasil.invalidate('CURRENT_ACCESS_TOKEN','CLIENT_TOKEN')
        ```
    """
    payload = {
        'accessToken': access_token,
        'clientToken': client_token
    }
    response = requests.post(URLs.invalidate(), json=payload)
    data = handle_response(response, PayloadError, TokenError)



import io
import socket
import struct
import time
from typing import IO, Optional, Tuple

from ._structures import ServerStats
from .packets import Packets


def read_null_terminated_string(buffer: IO, encoding: str = 'utf-8') -> str:
    res = b''
    char = buffer.read(1)
    while char != b'\0':
        res += char
        char = buffer.read(1)

    return res.decode(encoding)


def _parse_stats(data: bytes) -> ServerStats:
    with io.BytesIO(data) as buffer:
        # Read server info
        info = {}
        for _ in range(10):
            key = read_null_terminated_string(buffer)
            value = read_null_terminated_string(buffer)
            info[key] = value

        # TODO: Parse plugins
        info['motd'] = info.pop('hostname')
        info['game_type'] = info.pop('gametype')
        info['players'] = (int(info.pop('numplayers')), int(info.pop('maxplayers')))
        info['host'] = (info.pop('hostip'), int(info.pop('hostport')))
        
        # Skip next 11 bytes
        buffer.seek(11, 1)

        # Read players
        players = []
        player_name = read_null_terminated_string(buffer)
        while len(player_name) != 0:
            players.append(player_name)
            player_name = read_null_terminated_string(buffer)

    return ServerStats(**info, player_list=players)


def _handshake(sock: socket.socket, addr: Tuple[str, int], session_id: int) -> int:
    pcks = Packets(sock)
    pcks.send(9, session_id)

    r_type, r_session_id, data = pcks.recv()
    if r_type != 9 or r_session_id != session_id:
        raise Exception('An error occured while handshaking')

    return int(data.rstrip(b'\0'))


def _get_stats(sock: socket.socket, addr: Tuple[str, int], session_id: int, token: int) -> ServerStats:
    pcks = Packets(sock)
    pcks.send(0, session_id, struct.pack('>iI', token, 0xFFFFFF01))

    total_data = b''
    packet_id = 0
    while packet_id != 0x80:
        r_type, r_session_id, data = pcks.recv()
        packet_id = struct.unpack('>9xBx', data[:11])[0]

        if r_type != 0 or r_session_id != session_id:
            raise Exception('An error occured while getting stats')

        total_data += data[11:]

    return _parse_stats(total_data)


def get_stats(addr: Tuple[str, int], timeout: Optional[float] = 3) -> ServerStats:
    """Returns full stats about server using the Query protocol

    Args:
        addr (tuple): tuple with the address and the port to connect to
        timeout (int, optional): Time to wait before closing pending connection (default to 3)
    
    Returns:
        ServerStats

    Example:

        ```python
        from mojang.minecraft import query

        stats = query.get_stats(('localhost', 25585))
        print(stats)
        ```
        ```bash
        ServerStats(
            motd='A Minecraft Server', 
            game_type='SMP', 
            game_id='MINECRAFT', 
            version='1.16.5', 
            map='world', 
            host=('localhost', 25585), 
            players=(0, 20), 
            player_list=[]
        )
        ```

    """
    session_id = int(time.time()) & 0x0F0F0F0F

    try:
        with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
            sock.settimeout(timeout)
            sock.connect(addr)

            token = _handshake(sock, addr, session_id)
            stats = _get_stats(sock, addr, session_id, token)
    except socket.timeout:
        stats = None
    finally:
        return stats



import unittest

import mojang
from mojang.account.structures.base import NameInfoList, UserProfile, UUIDInfo, NameInfo
from mojang.account.structures.session import Cape, Skin


class TestMojangAPI(unittest.TestCase):

    def test_existent_uuid(self):
        self.assertEqual(mojang.get_uuid('Notch').uuid, '069a79f444e94726a5befca90e38aaf5')
        self.assertEqual(mojang.get_uuid('jeb_').uuid, '853c80ef3c3749fdaa49938b674adae6')
    
    def test_unexistent_uuid(self):
        self.assertEqual(mojang.get_uuid('UNEXISTENT_PLAYER'), None)

    def test_existent_uuids(self):
        self.assertEqual(mojang.get_uuids(['Notch','jeb_']), [UUIDInfo('Notch','069a79f444e94726a5befca90e38aaf5'), UUIDInfo('jeb_','853c80ef3c3749fdaa49938b674adae6')])
        self.assertEqual(mojang.get_uuids(['jeb_','Notch']), [UUIDInfo('jeb_','853c80ef3c3749fdaa49938b674adae6'), UUIDInfo('Notch','069a79f444e94726a5befca90e38aaf5')])

    def test_unexistent_uuids(self):
        self.assertEqual(mojang.get_uuids(['jeb_','UNEXISTENT_PLAYER']), [UUIDInfo('jeb_','853c80ef3c3749fdaa49938b674adae6'),None])
        self.assertEqual(mojang.get_uuids(['UNEXISTENT_PLAYER1','UNEXISTENT_PLAYER2']), [None,None])

    def test_existent_names(self):
        self.assertEqual(mojang.names('069a79f444e94726a5befca90e38aaf5'), NameInfoList([NameInfo('Notch', None)]))
        self.assertEqual(mojang.names('853c80ef3c3749fdaa49938b674adae6'), NameInfoList([NameInfo('jeb_', None)]))

    def test_unexistent_names(self):
        self.assertEqual(mojang.names('069a79f444e94726a5befca90e38aaf6'), NameInfoList([]))

    def test_existent_profile(self):
        self.assertEqual(mojang.user('069a79f444e94726a5befca90e38aaf5'), UserProfile('Notch', '069a79f444e94726a5befca90e38aaf5', False, False, NameInfoList([NameInfo('Notch', None)]), Skin('http://textures.minecraft.net/texture/292009a4925b58f02c77dadc3ecef07ea4c7472f64e0fdc32ce5522489362680', 'classic'), None))
        self.assertEqual(mojang.user('853c80ef3c3749fdaa49938b674adae6'), UserProfile('jeb_', '853c80ef3c3749fdaa49938b674adae6', False, False, NameInfoList([NameInfo('jeb_', None)]), Skin('http://textures.minecraft.net/texture/7fd9ba42a7c81eeea22f1524271ae85a8e045ce0af5a6ae16c6406ae917e68b5', 'classic'), Cape('http://textures.minecraft.net/texture/5786fe99be377dfb6858859f926c4dbc995751e91cee373468c5fbf4865e7151', None)))

    def test_unexistent_profile(self):
        self.assertEqual(mojang.user('069a79f444e94726a5befca90e38aaf6'), None)



import socket
import struct
from typing import Optional, Tuple


class Packets:
    
    def __init__(self, sock: socket.socket):
        self.__sock = sock

    def send(self, _type: int, sess_id: int, data: Optional[bytes] = b'') -> int:
        packet = struct.pack(f'>Hbi{len(data)}s', 0xFEFd, _type, sess_id, data)
        return self.__sock.send(packet)

    def recv(self) -> Tuple[int, int, bytes]:
        with self.__sock.makefile('rb') as buffer:
            data = buffer.read1()

        _type, sess_id = struct.unpack('>bi', data[:5])
        return _type, sess_id, data[5:]




class URLs:
    
    # Base URLs
    @classmethod
    def status_check(cls):
        """Returns the url to check Mojang APIs status"""
        return 'https://status.mojang.com/check'

    @classmethod
    def uuid(cls, username: str):
        """Returns the url to get uuid of username"""
        return f'https://api.mojang.com/users/profiles/minecraft/{username}'

    @classmethod
    def uuids(cls):
        """Returns the url to get uuids of multiple username"""
        return 'https://api.mojang.com/profiles/minecraft'

    @classmethod
    def name_history(cls, uuid: str):
        """Returns the url to get the name history of a user"""
        return f'https://api.mojang.com/user/profiles/{uuid}/names'

    @classmethod
    def profile(cls, uuid: str):
        """Returns the url to get the profile of a user"""
        return f'https://sessionserver.mojang.com/session/minecraft/profile/{uuid}'

    # Session URLs
    @classmethod
    def name_change(cls):
        """Returns the url to get t name change information"""
        return 'https://api.minecraftservices.com/minecraft/profile/namechange'
    
    @classmethod
    def change_name(cls, name: str):
        """Returns the url to change user's name"""
        return f'https://api.minecraftservices.com/minecraft/profile/name/{name}'
    
    @classmethod
    def change_skin(cls):
        """Returns the url to change the user's skin"""
        return 'https://api.minecraftservices.com/minecraft/profile/skins'
    
    @classmethod
    def reset_skin(cls, uuid: str):
        """Returns the url to reset the user's skin"""
        return f'https://api.mojang.com/user/profile/{uuid}/skin'



from typing import List, NamedTuple, Tuple


class Players(NamedTuple):
    count: Tuple[int, int]
    list: List[str]


class SLPResponse(NamedTuple):
    protocol_version: int
    version: str
    motd: str
    players: Players
    ping: float



import datetime as dt
import re
from os import path
from typing import NamedTuple
from urllib.parse import urlparse

import requests
import validators


class NameChange(NamedTuple):
    """
    Attributes:
        allowed (bool): Wether the user can change name
        created_at (dt.datetime): When was the user created
    """
    allowed: bool
    created_at: dt.datetime

class _SkinCapeBase(NamedTuple):
    source: str
    variant: str = None

    @classmethod
    def _filename_from_url(cls, url: str):
        url_path = urlparse(url).path
        match = re.match(r'^([\w,\s-]+)\.([A-Za-z]{3})$', path.basename(url_path))
        if match:
            return match.groups()

    @classmethod
    def _filename_from_headers(cls, headers: dict):
        # Check content-disposition
        if 'content-disposition' in headers.keys():
            cdisp = headers['content-disposition']
            file_names = re.findall('filename=(.+)', cdisp)
            if len(file_names) > 0:
                return file_names[0][0], file_names[0][1][1:]

        # Check content-type
        if 'content-type' in headers.keys():
            ctype = headers['content-type']
            if (not 'text' in ctype) and (not 'html' in ctype):
                return ctype.split('/')
    
    @classmethod
    def _download_bytes(cls, url: str):
        response = requests.get(url)
        if response.ok:
            filename = cls._filename_from_headers(response.headers) or cls._filename_from_url(url) or ['download', None]
            return filename, response.content

    @property
    def data(self):
        if not hasattr(self, '_data'):
            _data = b''
            _extension = None

            if validators.url(self.source):
                response = self._download_bytes(self.source)
                if response:
                    _extension = response[0][1]
                    _data = response[1]
            elif path.exists(self.source):
                basename = path.basename(self.source)
                _extension = path.splitext(basename)[1][1:]
                
                with open(self.source, 'rb') as fp:
                    _data = fp.read()

            if _extension != 'png':
                pass # TODO: Raise Exception
            
            object.__setattr__(self, '_data', _data)
        
        return self._data

    def save(self, dest: str):
        if not dest.endswith('.png'):
            dest += '.png'
        
        with open(dest, 'wb') as fp:
            fp.write(self.data)

class Skin(_SkinCapeBase):
    """
    Attributes:
        source (str): The source where the skin is located
        variant (str): The variant of skin (default to 'classic')
    """

class Cape(_SkinCapeBase):
    """
    Attributes:
        source (str): The source where the cape is located
    """



import json
import socket
import struct
import time
from typing import Optional

from .._structures import Players, SLPResponse
from .packets import Packets
from .types import String, VarInt


def ping(sock: socket.socket, hostname: Optional[str] = 'locahost', port: Optional[int] = 25565) -> SLPResponse:
    pcks = Packets(sock)

    # Send handshake packet    
    with pcks.send() as buffer:
        VarInt.write(buffer, 0x00)
        VarInt.write(buffer, 0)
        String.write(buffer, hostname)
        buffer.write(struct.pack('>H', port))
        VarInt.write(buffer, 0x01)

    # Send request packet
    with pcks.send() as buffer:
        VarInt.write(buffer, 0x00)
    
    # Receive response
    with pcks.recv() as buffer:
        _ = VarInt.read(buffer)
        response = json.loads(String.read(buffer))

    # Ping
    with pcks.send() as buffer:
        VarInt.write(buffer, 0x01)
        buffer.write(struct.pack('>q', int(time.time() * 1000)))
    
    # Pong
    with pcks.recv() as buffer:
        _ = VarInt.read(buffer)
        ping_start = struct.unpack('>q', buffer.read(8))[0]

    players = None
    if 'players' in response:
        players = Players(
            (response['players']['online'], response['players']['max']), 
            response['players'].get('sample', []))


    return SLPResponse(
        protocol_version=response.get('version', {}).get('protocol', 'unknown'), 
        version=response.get('version', {}).get('name', 'unknown'), 
        motd=response.get('description', {}).get('text', 'unknown'),
        players=players,
        ping=(time.time() * 1000) - ping_start
    )



import datetime as dt
from typing import Iterator, NamedTuple, Union

from .session import Cape, Skin


# Status check
class ServiceStatus(NamedTuple):
    """
    Attributes:
        name (str): The service name
        status (str): The service status
    """
    name: str
    status: str


class StatusCheck(tuple):
    
    def get(self, name: str) -> Union[None, 'ServiceStatus']:
        """Get service by name

        Args:
            name (str): The service name
        
        Returns:
            ServiceStatus
        """
        service = list(filter(lambda s: s.name == name, self))
        if len(service) > 0:
            return service[0]

    def __getitem__(self, x) -> ServiceStatus:
        return super().__getitem__(x)

    def __iter__(self) -> Iterator['ServiceStatus']:
        return super().__iter__()


# UUID and Name
class UUIDInfo(NamedTuple):
    """
    Attributes:
        name (str): The user name
        uuid (str): The user uuid
        legacy (bool): Wether the account has migrated
        demo (bool): Wether the account is demo
    """
    name: str
    uuid: str
    legacy: bool = False
    demo: bool = False


class NameInfo(NamedTuple):
    """
    Attributes:
        name (str): The player name
        changed_to_at (dt.datetime): When it's was changed to
    """
    name: str
    changed_to_at: dt.datetime


class NameInfoList(tuple):

    @property
    def current(self) -> NameInfo:
        """Returns the most recent name"""
        if len(self) == 1:
            return self[0]

        _list = filter(lambda n: n.change_to_at != None, self)
        return max(_list, key=lambda n: n.change_to_at)
    
    @property
    def first(self) -> NameInfo:
        """Returns the first name"""
        first = list(filter(lambda n: n.changed_to_at == None, self))
        return first[0]

    def __getitem__(self, x) -> NameInfo:
        return super().__getitem__(x)

    def __iter__(self) -> Iterator['NameInfo']:
        return super().__iter__()


## Profile
class UserProfile(NamedTuple):
    """
    Attributes:
        name (str): The user name
        uuid (str): The user uuid
        is_legacy (bool): Wether the account has migrated
        is_demo (bool): Wether the account is demo
        names (NameInfoList): The user name history
        skin (Skin): The user skin
        cape (Cape): The user cape
    """
    name: str
    uuid: str
    is_legacy: bool
    is_demo: bool
    names: NameInfoList
    skin: Skin
    cape: Cape



import datetime as dt

import requests

from ..exceptions import *
from .structures.session import NameChange, Skin
from .utils.auth import BearerAuth
from .utils.urls import URLs


def get_user_name_change(access_token: str) -> NameChange:
    """Return if user can change name and when it was created
    
    Args:
        access_token (str): The session's access token
    
    Returns:
        NameChange
    
    Raises:
        Unauthorized: If the access token is invalid

    Example:

        ```python
        from mojang.account import session

        name_change = session.get_user_name_change('ACCESS_TOKEN')
        print(name_change)
        ```
        ```bash
        NameChange(allowed=True, created_at=datetime.datetime(2006, 4, 29, 10, 10, 10))
        ```
    """
    response = requests.get(URLs.name_change(), auth=BearerAuth(access_token))
    data = handle_response(response, PayloadError, Unauthorized)

    data['created_at'] = dt.datetime.strptime(data.pop('createdAt'), '%Y-%m-%dT%H:%M:%SZ')
    data['allowed'] = data.pop('nameChangeAllowed')

    return NameChange(**data)

def change_user_name(access_token: str, name: str):
    """Change name of authenticated user
    
    Args:
        access_token (str): The session's access token
        name (str): The new user name
    
    Raises:
        Unauthorized: If the access token is invalid
        InvalidName: If the new user name is invalid
        UnavailableName: If the new user name is unavailable
    
    Example:

        ```python
        from mojang.account import session

        session.change_user_name('ACCESS_TOKEN', 'my_super_cool_name')
        ```
    """
    response = requests.put(URLs.change_name(name), auth=BearerAuth(access_token))
    handle_response(response, InvalidName, UnavailableName, Unauthorized)

def change_user_skin(access_token: str, path: str, variant='classic'):
    """Change skin of authenticated user
    
    Args:
        access_token (str): The session's access token
        path (str): The the path to the new skin, either local or remote
        variant (str, optional): The skin variant, either `classic` or `slim`
    
    Raises:
        Unauthorized: If the access token is invalid

    Example:

        ```python
        from mojang.account import session

        session.change_user_skin('ACCESS_TOKEN', 'http://...')
        ```
    """
    skin = Skin(source=path, variant=variant)
    files = [
        ('variant', skin.variant),
        ('file', ('image.png', skin.data, f'image/png'))
    ]
    response = requests.post(URLs.change_skin(), auth=BearerAuth(access_token), files=files, headers={'content-type': None})
    handle_response(response, PayloadError, Unauthorized)

def reset_user_skin(access_token: str, uuid: str):
    """Reset skin of authenticated user
    
    Args:
        access_token (str): The session's access token
        uuid (str): The user uuid
    
    Raises:
        Unauthorized: If the access token is invalid

    Example:

        ```python
        from mojang.account import session

        session.reset_user_skin('ACCESS_TOKEN', 'USER_UUID')
        ```
    """
    response = requests.delete(URLs.reset_skin(uuid), auth=BearerAuth(access_token))
    handle_response(response, PayloadError, Unauthorized)



import base64
import json
from typing import List

import requests

from ..exceptions import handle_response
from .structures.base import *
from .utils.urls import URLs


def status() -> StatusCheck:
    """Get the status of Mojang's services
    
    Returns:
        StatusCheck
    
    Example:

        Get status for all services
        ```python
        import mojang

        status = mojang.status()
        print(status)
        ```
        ```bash
        (
            ServiceStatus(name='minecraft.net', status='green'), 
            ServiceStatus(name='session.minecraft.net', status='green'), 
            ServiceStatus(name='account.mojang.com', status='green'), 
            ServiceStatus(name='authserver.mojang.com', status='green'), 
            ServiceStatus(name='sessionserver.mojang.com', status='red'),
            ServiceStatus(name='api.mojang.com', status='green'), 
            ServiceStatus(name='textures.minecraft.net', status='green'), 
            ServiceStatus(name='mojang.com', status='green')
        )
        ```

        Get status for one specific service
        ```python
        import mojang

        status = mojang.status().get('minecraft.net')
        print(status)
        ```
        ```bash
        ServiceStatus(name='minecraft.net', status='green')
        ```
    """
    response = requests.get(URLs.status_check())
    data = handle_response(response)

    _status = []
    for service in data:
        item = list(service.items())[0]
        _status.append(ServiceStatus(name=item[0], status=item[1]))
    
    return StatusCheck(_status)

def get_uuid(username: str) -> UUIDInfo:
    """Get uuid of username
    
    Args:
        username (str): The username which you want the uuid of
    
    Returns:
        UUIDInfo

    Example:

        ```python
        import mojang
        uuid_info = mojang.get_uuid('Notch')
        print(uuid_info)
        ```
        ```
        UUIDInfo(name='Notch', uuid='069a79f444e94726a5befca90e38aaf5', legacy=False, demo=False)
        ```
    """
    response = requests.get(URLs.uuid(username))
    data = handle_response(response)

    if not data:
        return None

    data['uuid'] = data.pop('id')

    return UUIDInfo(**data)

def get_uuids(usernames: list) -> List['UUIDInfo']:
    """Get uuid of multiple username
    
    Note: Limited Endpoint
        The Mojang API only allow 10 usernames maximum, if more than 10 usernames are
        given to the function, multiple request will be made.
    
    Args:
        usernames (list): The list of username which you want the uuid of
    
    Returns:
        A list of UUIDInfo

    Example:

        ```python
        import mojang
        uuids_info = mojang.get_uuids(['Notch', '_jeb'])
        print(uuids_info)
        ```
        ```
        [
            UUIDInfo(name='Notch', uuid='069a79f444e94726a5befca90e38aaf5', legacy=False, demo=False), 
            UUIDInfo(name='_jeb', uuid='45f50155c09f4fdcb5cee30af2ebd1f0', legacy=False, demo=False)
        ]
        ```
    """
    usernames = list(map(lambda u: u.lower(), usernames))
    _uuids = [None]*len(usernames)

    for i in range(0, len(usernames), 10):
        response = requests.post(URLs.uuids(), json=usernames[i:i+10])
        data = handle_response(response)

        for item in data:
            index = usernames.index(item['name'].lower())
            item['uuid'] = item.pop('id')
            _uuids[index] = UUIDInfo(**item)

    return _uuids

def names(uuid: str) -> NameInfoList:
    """Get the user's name history
    
    Args:
        uuid (str): The user's uuid
    
    Returns:
        NameInfoList

    Example:

        ```python
        import mojang

        name_history = mojang.names('65a8dd127668422e99c2383a07656f7a')
        print(name_history)
        ```
        ```
        (
            NameInfo(name='piewdipie', changed_to_at=None), 
            NameInfo(name='KOtMotros', changed_to_at=datetime.datetime(2020, 3, 4, 17, 45, 26))
        )
        ```
    """
    response = requests.get(URLs.name_history(uuid))
    data = handle_response(response)

    _names = []
    for item in data:
        changed_to_at = None
        if 'changedToAt' in item.keys():
            changed_to_at = dt.datetime.fromtimestamp(item['changedToAt'] / 1000)
        _names.append(NameInfo(name=item['name'], changed_to_at=changed_to_at))

    return NameInfoList(_names)

def user(uuid: str) -> UserProfile:
    """Returns the full profile of a user 
    
    Args:
        uuid (str): The uuid of the profile
    
    Returns:
        UserProfile

    Example:

        ```python
        import mojang

        profile = mojang.user('069a79f444e94726a5befca90e38aaf5')
        print(profile)
        ```
        ```
        UserProfile(
            name='Notch', 
            uuid='069a79f444e94726a5befca90e38aaf5', 
            is_legacy=False, 
            is_demo=False, 
            names=(NameInfo(name='Notch', changed_to_at=None),),
            skin=Skin(source='...', variant='classic'), 
            cape=None
        )
        ```
    """
    response = requests.get(URLs.profile(uuid))
    data = handle_response(response)
    
    if not data:
        return None

    _dict = dict.fromkeys(UserProfile._fields, None)

    # Load profile info
    _dict['name'] = data['name']
    _dict['uuid'] = uuid
    _dict['is_legacy'] = data.get('legacy', False)
    _dict['is_demo'] = data.get('demo', False)

    # Load skin and cape
    textures_data = json.loads(base64.b64decode(data['properties'][0]['value']))
    
    skin_data = textures_data['textures'].get('SKIN', None)
    if skin_data:
        _dict['skin'] = Skin(skin_data['url'], skin_data.get('metadata', {'model': 'classic'})['model']) 

    cape_data = textures_data['textures'].get('CAPE', None)
    if cape_data:
        _dict['cape'] = Cape(cape_data['url'])

    # Get name history
    _dict['names'] = names(uuid)

    return UserProfile(**_dict)



import socket
from contextlib import contextmanager
from typing import Callable, Optional, Tuple

from .packets import Packets


@contextmanager
def session(addr: Tuple[str, int], password: str, timeout: Optional[float] = 3) -> Callable[[str], str]:
    """Open a RCON connection

    Args:
        addr (tuple): The address and the port to connect to
        password (str): The RCON password set in the server properties
        timeout (int, optional): Time to wait before closing pending connection (default to 3) 

    Returns:
        A function to send command
        
    Example:

        ```python
        from mojang.minecraft import rcon

        with rcon.session(('localhost', 25575), 'my_super_password') as send:
            result = send('help') # This execute the /help command
        ```

    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(timeout)
    sock.connect(addr)
    pcks = Packets(sock)

    packet_id = pcks.send(3, password)[0]
    r_packet_id, r_type = pcks.recv()[:2]

    # Check packet id and packet type
    if r_packet_id != packet_id or r_type != 2:
        raise Exception('Authentication failed')

    def send(command: str):
        # TODO: Parse command response
        packet_id = pcks.send(2, command)[0]
        r_packet_id, r_type, payload, size = pcks.recv()

        # Check packet id and packet type
        if r_packet_id != packet_id or r_type != 0:
            raise Exception('Command error')

        return payload.strip(b'\0').decode('ascii')

    try:
        yield send
    finally:
        sock.close()



import datetime as dt
from typing import Optional, Tuple

from ..structures.session import Cape, Skin

from .. import session, user
from ..auth import security, yggdrasil
from ..structures.base import NameInfoList


def connect(username: str, password: str, client_token: Optional[str] = None) -> 'UserSession':
    """Connect a user with name and password
    
    Args:
        username (str): The username or email if account is not legacy
        password (str): The user password
        client_token (str, optional): The client token to use in the authentication (default to None)
    
    Returns:
        UserSession

    Example:

        ```python
        import mojang

        session = mojang.connect('USERNAME_OR_EMAIL', 'PASSWORD')
        print(session)
        ```
        ```bash
        UserSession(
            name='PLAYER_NAME', 
            uuid='PLAYER_UUID', 
            is_legacy=False, 
            is_demo=False, 
            names=(NameInfo(name='PLAYER_NAME', changed_to_at=None),), 
            skin=Skin(source='http://...', variant='classic'), 
            cape=None, 
            created_at=datetime.datetime(2006, 4, 29, 10, 10, 10), 
            name_change_allowed=True
        )
        ```
    """
    auth = yggdrasil.authenticate(username, password, client_token)
    return UserSession(auth.access_token, auth.client_token)



class UserSession:
    """
    Attributes:
        name (str): The user name
        uuid (str): The user uuid
        is_legacy (bool): Wether the account has migrated
        is_demo (bool): Wether the account is demo
        names (NameInfoList): The user name history
        skin (Skin): The user skin
        cape (Cape): The user cape
        name_change_allowed (bool): Wether the user can change name
        created_at (dt.datetime): When was the user created
    """
    name: str
    uuid: str
    is_legacy: bool
    is_demo: bool
    names: NameInfoList
    skin: Skin
    cape: Cape
    created_at: dt.datetime
    name_change_allowed: bool

    def __init__(self, access_token: str, client_token: str):
        """Create a user session with access token and client token.
        The access token will be refreshed once the class is initiated

        Args:
            access_token (str): The session's access token
            client_token (str): The session's client token
        """
        self.__access_token = access_token
        self.__client_token = client_token

        self.refresh()

    def refresh(self):
        """Refresh the full user session, including the data"""
        auth = yggdrasil.refresh(self.__access_token, self.__client_token)

        # Update tokens
        self.__access_token = auth.access_token
        self.__client_token = auth.client_token

        # Update info
        self.uuid = auth.uuid
        self.name = auth.name
        self.is_demo = auth.demo
        self.is_legacy = auth.legacy

        # Fetch other data
        self._fetch_data()

    def _fetch_data(self):
        # Load profile
        profile = user(self.uuid)
        self.names = profile.names
        self.skin = profile.skin
        self.cape = profile.cape
        del profile

        # Load name change
        name_change = session.get_user_name_change(self.__access_token)
        self.name_change_allowed = name_change.allowed
        self.created_at = name_change.created_at
    
    def close(self):
        """Close the session and invalidates the access token"""
        yggdrasil.invalidate(self.__access_token, self.__client_token)
        self.__access_token = None
        self.__client_token = None

    @property
    def token_pair(self) -> Tuple[str, str]:
        """Returns the access token and the client token"""
        return self.__access_token, self.__client_token

    # Security
    @property
    def secure(self):
        """Check wether user IP is secured. For more details checkout [`check_ip`][mojang.account.auth.security.check_ip]"""
        return security.check_ip(self.__access_token)

    @property
    def challenges(self):
        """Returns the list of challenges to verify user IP. For more details checkout [`get_challenges`][mojang.account.auth.security.get_challenges]"""
        return security.get_challenges(self.__access_token)

    def verify(self, answers: list):
        """Verify user IP. For more details checkout [`verify_ip`][mojang.account.auth.security.verify_ip]"""
        return security.verify_ip(self.__access_token, answers)

    # Name
    def change_name(self, name: str):
        """Change user name. For more details checkout [`change_user_name`][mojang.account.session.change_user_name]
        
        Args:
            name (str): The new name
        """
        session.change_user_name(self.__access_token, name)
        self._fetch_data()

    # Skin
    def change_skin(self, path: str, variant: Optional[str] = 'classic'):
        """Change user skin. For more details checkout [`change_user_skin`][mojang.account.session.change_user_skin]

        Args:
            path (str): The path to the skin, either local or remote
            variant (str, optional): The variant of skin (default to 'classic')
        """
        session.change_user_skin(self.__access_token, path, variant)
        self._fetch_data()

    def reset_skin(self):
        """Reset user skin. For more details checkout [`reset_user_skin`][mojang.account.session.reset_user_skin]"""
        session.reset_user_skin(self.__access_token, self.uuid)
        self._fetch_data()



import io
import socket
from contextlib import contextmanager

from .types import VarInt


class Packets:

    def __init__(self, sock: socket.socket):
        self.__sock = sock


    @contextmanager
    def send(self):
        packet = io.BytesIO()
        try:
            yield packet
        finally:
            data = packet.getvalue()
            with self.__sock.makefile('wb') as buffer:
                VarInt.write(buffer, len(data))
                buffer.write(data)


    def recv(self):
        with self.__sock.makefile('rb') as buffer:
            length = VarInt.read(buffer)
            packet = buffer.read(length)

        return io.BytesIO(packet)



import socket
from functools import partial
from typing import Optional, Tuple
import enum

from ._structures import SLPResponse
from .post_netty import ping as mping
from .pre_netty import ping_fe, ping_fe01

class PingVersion(enum.IntFlag):
    """
    Attributes:
        V1_7 (int): Use the ping protocol for version 1.7 and higher
        V1_6 (int): Use the ping protocol for version 1.6
        V1_4 (int): Use the ping protocol for version 1.4 and 1.5
        V1_3 (int): Use the ping protocol for version 1.3 and lower
        V_ALL (int): Use all the ping protocol
    """
    V1_7 = 1
    V1_6 = 2
    V1_4 = 4
    V1_3 = 8
    V_ALL = 15


def ping(addr: Tuple[str, int], timeout: Optional[int] = 3, flags: Optional[int] = PingVersion.V_ALL) -> SLPResponse:
    """Ping the server for information

    Args:
        addr (tuple): The address and the port to connect to
        timeout (int, optional): Time to wait before closing pending connection (default to 3) 
        flags (int, optional): Which version of the ping version to use

    Returns:
        SLPResponse

    Example:

        ```python
        from mojang.minecraft import slp

        stats = slp.ping(('localhost', 25565))
        print(stats)
        ```
        ```bash
        SLPResponse(
            protocol_version=754,
            version='1.16.5', 
            motd='A Minecraft Server', 
            players=Players(count=(0, 20), list=[]),
            ping=1
        )
        ```

    """
    _methods = []
    if PingVersion.V1_7 & flags:
        _methods.append(partial(mping, hostname=addr[0], port=addr[1]))
    if PingVersion.V1_6 & flags:
        _methods.append(partial(ping_fe01, hostname=addr[0], port=addr[1]))
    if PingVersion.V1_4 & flags:
        _methods.append(ping_fe01)
    if PingVersion.V1_3 & flags:
        _methods.append(ping_fe)

    response = None
    while len(_methods) > 0 and response is None:
        method = _methods.pop(0)
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.settimeout(timeout)
            try:
                sock.connect(addr)
                response = method(sock)
            except socket.error:
                pass
    
    return response
        


from typing import List

import requests

from ...exceptions import *
from ..structures.auth import ChallengeInfo
from ..utils.auth import BearerAuth, URLs


def check_ip(access_token: str) -> bool:
    """Check if authenticated user IP is secure
    
    Args:
        access_token (str): The session's access token
    
    Returns:
        True if IP is secure else False
    
    Raises:
        Unauthorized: If access token is invalid
        PayloadError: If access token is not formated correctly

    Example:

        ```python
        from mojang.account.auth import security

        checked = security.check_ip('ACCESS_TOKEN')
        print(checked)
        ```
        ```
        True
        ```
    """
    response = requests.get(URLs.verify_ip(), auth=BearerAuth(access_token))
    try:
        handle_response(response, PayloadError, Unauthorized, IPNotSecured)
    except IPNotSecured:
        return False
    else:
        return True

def get_challenges(access_token: str) -> List['ChallengeInfo']:
    """Return a list of challenges to verify IP
    
    Args:
        access_token (str): The session's access token
    
    Returns:
        A list of ChallengeInfo
    
    
    Raises:
        Unauthorized: If access token is invalid
        PayloadError: If access token is not formated correctly

    Example:
        
        ```python
        from mojang.account.auth import security

        challenges = security.get_challenges('ACCESS_TOKEN')
        print(challenges)
        ```
        ```bash
        [
            ChallengeInfo(id=123, challenge="What is your favorite pet's name?"), 
            ChallengeInfo(id=456, challenge="What is your favorite movie?"), 
            ChallengeInfo(id=589, challenge="What is your favorite author's last name?")
        ]
        ```
    """
    response = requests.get(URLs.get_challenges(), auth=BearerAuth(access_token))
    data = handle_response(response, PayloadError, Unauthorized)

    _challenges = []
    for item in data:
        _challenges.append(ChallengeInfo(id=item['answer']['id'], challenge=item['question']['question']))

    return _challenges

def verify_ip(access_token: str, answers: list) -> bool:
    """Verify IP with the given answers
    
    Args:
        access_token (str): The session's access token
        answers (list): The answers to the question
        
    Returns:
        True if IP is secure else False
    
    Raises:
        Unauthorized: If access token is invalid
        PayloadError: If access token is not formated correctly
    
    Example:
        
        ```python
        from mojang.account.auth import security

        answers = [
            (123, "foo"),
            (456, "bar"),
            (789, "baz")
        ]

        security.verify_user_ip('ACCESS_TOKEN', answers)
        ```
    """
    answers = list(map(lambda a: {'id': a[0], 'answer': a[1]}, answers))
    response = requests.post(URLs.verify_ip(), auth=BearerAuth(access_token), json=answers)
    try:
        handle_response(response, PayloadError, Unauthorized, IPVerificationError)
    except IPVerificationError:
        return False
    else:
        return True



import setuptools

with open('./README.md', 'r') as fp:
    long_description = fp.read()

setuptools.setup(
    name='pymojang',
    version='1.2.1',
    author='Lucino772',
    author_email='lucapalmi772@gmail.com',
    licence='MIT',
    url='https://github.com/Lucino772/pymojang',
    description='It\'s a full wrapper arround the \
         Mojang API and Mojang Authentication API',
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9"
    ],
    packages=setuptools.find_packages(),
    install_requires=['requests', 'validators'],
    keywords=['minecraft', 'mojang', 'python3'],
    project_urls={
        'Documentation': 'https://pymojang.readthedocs.io/en/latest/'
    }
)




def handle_response(response, *exceptions, use_defaults=True):
    """Handle response message from http request. Every given `exception`
    must have a `code` property.

    The function will check if the status code from the response is ok.
    If not an Exception will be raised based on the status code.
    """
    if response.ok:
        data = {}
        try:
            data = response.json()
        except ValueError:
            pass
        finally:
            return data
    else:
        if use_defaults:
            exceptions += (NotFound, MethodNotAllowed, ServerError)
        data = response.json()
        for exception in exceptions:
            if isinstance(exception.code, int):
                if response.status_code == exception.code:
                    raise exception(data['errorMessage'])
            elif isinstance(exception.code, list):
                if response.status_code in exception.code:
                    raise exception(data['errorMessage'])
        else:
            raise Exception(*data.values())


# Global
class MethodNotAllowed(Exception):
    """The method used for the request is not allowed"""
    code = 405


class NotFound(Exception):
    """The requested url doesn't exists"""
    code = 404


class ServerError(Exception):
    """There is an internal error on the server"""
    code = 500


class PayloadError(Exception):
    """The data sent to the server has an invalid format"""
    code = 400


# Authentication Errors
class CredentialsError(Exception):
    """The credentials sent to the server are wrong"""
    code = [403, 429]


class TokenError(Exception):
    """The token sent to the server has an invalid format"""
    code = 403


class Unauthorized(Exception):
    """The token sent to the server is invalid"""
    code = 401


# Name Change Errors
class InvalidName(Exception):
    """The name is invalid, longer than 16 characters or contains
    characters other than (a-zA-Z0-9_). Only raised when changing
    the name of a user
    """
    code = 400

    def __init__(self):
        super().__init__("Name is invalid, longer than 16 characters or \
             contains characters other than (a-zA-Z0-9_)")


class UnavailableName(Exception):
    """Name is unavailable. Only raised when changing the name of a user"""
    code = 403

    def __init__(self):
        super().__init__("Name is unavailable")


# Security
class IPNotSecured(Exception):
    """The current IP is not secured. Only raised when
    checking if user IP is secure
    """
    code = 403


class IPVerificationError(Exception):
    """Verifiction for IP failed. Only raised when verifying user IP"""
    code = 403



from typing import List, NamedTuple, Tuple


class ServerStats(NamedTuple):
    motd: str
    game_type: str
    game_id: str
    version: str
    map: str
    host: Tuple[str, int]
    plugins: List[str]
    players: Tuple[int, int]
    player_list: List[str]



import socket
import struct
import time
from typing import Optional

from .._structures import Players, SLPResponse


def ping_fe01(sock: socket.socket, hostname: Optional[str] = None, port: Optional[int] = -1):
    if hostname is not None and len(hostname) > 0 and port > 0:
        encoded_hostname = hostname.encode('utf-16be')
        packet = struct.pack(
            f'>BBBh22shBh{len(encoded_hostname)}si',
            0xFE, 0x01, 0xFA,                           # Packet info
            11, 'MC|PingHost'.encode('utf-16be'),       # MC|Pingost
            7 + len(encoded_hostname),                  # Length of data
            80, len(hostname), encoded_hostname, port   # Data
        )
    else:
        packet = struct.pack('>BB', 0xFE, 0x01)

    start = time.time() * 1000
    sock.send(packet)

    with sock.makefile('rb') as buffer:
        buffer.read1(9) # Skip 9 first bytes
        resp = buffer.read().decode('utf-16-be').split('\x00')
    
    latency = (time.time() * 1000) - start

    return SLPResponse(
        protocol_version=resp[0], 
        version=resp[1], 
        motd=resp[2],
        players=Players((int(resp[3]), int(resp[4])), []),
        ping=latency
    )


def ping_fe(sock: socket.socket):
    start = time.time() * 1000
    sock.send(struct.pack('>B', 0xFE))

    with sock.makefile('rb') as buffer:
        buffer.read(3) # Skip 3 first bytes
        resp = buffer.read().decode('utf-16-be').split('\xa7')
    
    latency = (time.time() * 1000) - start

    return SLPResponse(
        protocol_version=None, 
        version=None, 
        motd=resp[0],
        players=Players((int(resp[1]), int(resp[2])), []),
        ping=latency
    )
